(** * A small theory of non decreasing piecewise constant functions
    with finitely many pieces and left continuous. *)

Require Import mathcomp.ssreflect.ssreflect
        mathcomp.ssreflect.ssrbool mathcomp.ssreflect.ssrnat
        mathcomp.ssreflect.ssrfun mathcomp.ssreflect.eqtype
        mathcomp.ssreflect.seq mathcomp.ssreflect.path mathcomp.ssreflect.order.
Require Import Flocq.Core.Raux.
Require Import Reals mathcomp.analysis.Rstruct nnR.
Require Import Psatz.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Import Order.Theory.

Local Open Scope R_scope.

(** Consider a function [f] such that [f(x)] is [y_0] for all [x <=
    x_0], [f(t) = y_i] for all [t] in interval ( [x_{i-1}], [x_i] ]
    and [f(t) = c] for any [t > x_n]. This function can be represented
    by the sequence of pairs [s = [(x_0, y_0),..., (x_n, y_n)]] and
    the constant [c] where the sequence [[x_0,..., x_n]] is sorted, as
    well as the sequence [[y_0,..., y_n, c]] when f is non
    decreasing. *)
Record PWC := {
  PWC_s : seq (R * R);
  PWC_c : R;
  _ : sorted Rleb (unzip1 PWC_s);
  _ : sorted Rleb (rcons (unzip2 PWC_s) PWC_c);
}.

Definition PWC2f (p : PWC) (x : R) :=
  nth (PWC_c p) (unzip2 (PWC_s p)) (find (Rleb x) (unzip1 (PWC_s p))).

Lemma PWC_non_decr (f : PWC) : forall x y, x <= y -> PWC2f f x <= PWC2f f y.
Proof.
move=> x y Hxy.
case: f => s c Hs1 Hs2.
rewrite /PWC2f /=.
set ix := find (Rleb x) _.
set iy := find (Rleb y) _.
rewrite -[in X in _ <= X]nth_rcons_default -nth_rcons_default.
apply /RlebP.
apply sorted_leq_nth.
{ exact: le_trans. }
{ move=> x'; apply /RleP /Rle_refl. }
{ exact Hs2. }
{ rewrite size_rcons size_map.
  rewrite (_ : size s = size (unzip1 s)); [|by rewrite size_map].
  apply find_size. }
{ rewrite size_rcons size_map.
  rewrite (_ : size s = size (unzip1 s)); [|by rewrite size_map].
  apply find_size. }
apply sub_find => z /RleP Hyz.
apply /RleP.
by move: Hyz; apply Rle_trans.
Qed.

Lemma PWC_bound (f : PWC) : forall x, PWC2f f x <= PWC_c f.
Proof.
move=> x.
case: f => s c Hs1 Hs2 /=.
rewrite /PWC2f /=.
set ix := find (Rleb x) _.
rewrite [in X in _ <= X](_ : c = nth c (unzip2 s) (size s));
  [|by rewrite nth_default ?size_map].
rewrite -[in X in _ <= X]nth_rcons_default -nth_rcons_default.
apply /RlebP.
apply sorted_leq_nth.
{ exact: le_trans. }
{ move=> x'; apply /RleP /Rle_refl. }
{ exact Hs2. }
{ rewrite size_rcons size_map.
  rewrite (_ : size s = size (unzip1 s)); [|by rewrite size_map].
  apply find_size. }
{ by rewrite size_rcons size_map /in_mem /=. }
rewrite (_ : size s = size (unzip1 s)); [|by rewrite size_map].
apply find_size.
Qed.

Definition next x s := nth x s (find (Rleb x) s).

Lemma next_ge_id x s : x <= next x s.
Proof.
rewrite /next.
move: (find_size (Rleb x) s).
rewrite leq_eqVlt => /orP [/eqP|] Hfx.
{ rewrite nth_default ?Hfx //; apply Rle_refl. }
apply /RlebP /nth_find.
by rewrite has_find.
Qed.

Lemma PWC_step (f : PWC) (x : R) :
  forall y, x <= y <= next x (unzip1 (PWC_s f)) -> PWC2f f x = PWC2f f y.
Proof.
move=> y.
rewrite /PWC2f /next.
set s := unzip1 _.
move=> [Hy1 Hy2].
apply f_equal.
apply /anti_leq /andP; split.
{ apply sub_find => z /RleP Hz.
  by apply /RleP; move: Hz; apply Rle_trans. }
rewrite leqNgt.
apply /negP => H.
by move: Hy2 (before_find x H) => /RlebP ->.
Qed.

Definition in_segment (f : PWC) (x x' : R) :=
  all (fun t => Rleb x t && Rltb t x') (unzip1 (PWC_s f)).

Lemma PWC_left_segment f x x' :
  in_segment f x x' -> forall t, t <= x -> PWC2f f t = PWC2f f x.
Proof.
rewrite /in_segment /PWC2f => /allP Hxx' t Ht; move: Hxx'.
case f => s c _ _ /=.
case s => [//|[hs1 hs2] ts] /= Hxx'.
move: (Hxx' _ (mem_head _ _)) => /andP [Hxhs1 _].
have -> : Rleb t hs1.
{ apply /RleP; move: Hxhs1 => /RleP; lra. }
by rewrite Hxhs1.
Qed.

Lemma PWC_right_segment f x x' :
  in_segment f x x' -> forall t, x' <= t -> PWC2f f t = PWC2f f x'.
Proof.
rewrite /in_segment /PWC2f => /allP Hxx' t Ht; move: Hxx'.
case f => s c _ _ /=.
elim: s => [//|[hs1 hs2] ts IHs] /= Hxx'.
move: (Hxx' _ (mem_head _ _)) => /andP [_ Hxhs1].
have -> /= : Rleb t hs1 = false.
{ apply /RleP /Rlt_not_le; move: Hxhs1 => /RltP; lra. }
have -> /= : Rleb x' hs1 = false.
{ by apply /RleP /Rlt_not_le; move: Hxhs1 => /RltP. }
by apply IHs => t' Ht'; apply Hxx'; rewrite in_cons Ht' orbC.
Qed.

Program Definition PWC_cst (c : R) : PWC := @Build_PWC nil c _ _.

Program Definition PWC_singleton (x y y' : R) (Hyy' : y <= y') : PWC :=
  @Build_PWC [:: (x, y)] y' _ _.
Next Obligation. by rewrite andbC /=; apply /RleP. Qed.

Lemma PWC_singleton_before (x y y' : R) (Hyy' : y <= y') :
  forall x', x' <= x -> PWC2f (PWC_singleton x Hyy') x' = y.
Proof. move=> x' Hx'; cbv; by case Rle_dec; [|lra]. Qed.

Lemma PWC_singleton_after (x y y' : R) (Hyy' : y <= y') :
  forall x', x < x' -> PWC2f (PWC_singleton x Hyy') x' = y'.
Proof. move=> x' Hx'; cbv; by case Rle_dec; [lra|]. Qed.

Program Definition PWC_add (f g : PWC) : PWC :=
  @Build_PWC
    (let xs := merge Rleb (unzip1 (PWC_s f)) (unzip1 (PWC_s g)) in
     [seq (x, PWC2f f x + PWC2f g x)  | x <- xs])
    (PWC_c f + PWC_c g) _ _.
Next Obligation.
rewrite /unzip1 -map_comp map_id.
apply merge_sorted.
{ exact: le_total. }
{ by case f. }
by case g.
Qed.
Next Obligation.
pose xfl := head 0 (unzip1 (PWC_s f)).
pose xgl := head 0 (unzip1 (PWC_s g)).
pose xl := Rmin xfl xgl.
apply (path_sorted (x:=(0, PWC2f f xl + PWC2f g xl).2)).
rewrite rcons_path.
rewrite /unzip2 -map_comp.
set f' := snd \o _.
rewrite -[snd _]/(f' xl).
apply /andP; split; last first.
{ rewrite last_map /f' /=.
  by apply /RleP /Rplus_le_compat; apply PWC_bound. }
apply (@homo_path _ _ _ Rleb).
{ move=> x y /RleP Hxy.
  rewrite /f' /=.
  by apply /RleP /Rplus_le_compat; apply PWC_non_decr. }
apply merge_path.
{ exact: le_total. }
{ rewrite {}/xl {}/xfl.
  case: f {f'} => s _ Hs _ /=.
  move: Hs.
  case (unzip1 s) => [//|r l /= ->].
  rewrite andbC /=.
  apply /RleP /Rmin_l. }
rewrite {}/xl {}/xgl.
case: g {f'} => s _ Hs _ /=.
move: Hs.
case (unzip1 s) => [//|r l /= ->].
rewrite andbC /=.
apply /RleP /Rmin_r.
Qed.

Lemma PWC_add_correct (f g : PWC) :
  forall x, PWC2f (PWC_add f g) x = PWC2f f x + PWC2f g x.
Proof.
move=> x.
rewrite /PWC_add {1}/PWC2f /=.
set xfs := unzip1 (PWC_s f).
set xgs := unzip1 (PWC_s g).
set xs := merge _ xfs xgs.
rewrite /unzip1 /unzip2 -!map_comp map_id.
rewrite -[snd \o _]/(fun x' => PWC2f f x' + PWC2f g x').
move: (find_size (Rleb x) xs).
rewrite leq_eqVlt => /orP [/eqP|] Hfx.
{ rewrite nth_default ?size_map ?Hfx //.
  rewrite /PWC2f !nth_default // size_map.
  { rewrite leqNgt.
    rewrite (_ : size _ = size (unzip1 (PWC_s g))); [|by rewrite size_map].
    rewrite -has_find.
    apply /hasPn => x' Hx'.
    apply /negP => Hxx'.
    suff : has (Rleb x) xs; [by rewrite has_find Hfx ltnn|].
    apply /hasP; exists x'; [|exact Hxx'].
    rewrite mem_merge mem_cat.
    by apply /orP; right. }
  rewrite leqNgt.
  rewrite (_ : size _ = size (unzip1 (PWC_s f))); [|by rewrite size_map].
  rewrite -has_find.
  apply /hasPn => x' Hx'.
  apply /negP => Hxx'.
  suff : has (Rleb x) xs; [by rewrite has_find Hfx ltnn|].
  apply /hasP; exists x'; [|exact Hxx'].
  rewrite mem_merge mem_cat.
  by apply /orP; left. }
rewrite (nth_map x _ _ Hfx).
have Hsxs : sorted Rleb xs.
{ apply merge_sorted.
  { exact: le_total. }
  { by rewrite /xfs; case f. }
  by rewrite /xgs; case g. }
apply f_equal2; symmetry.
{ case_eq (has (Rleb x) xfs) => Hhas; last first.
  { apply Rle_antisym.
    { apply /PWC_non_decr /next_ge_id. }
    rewrite [in X in _ <= X]/PWC2f [in X in _ <= X]nth_default.
    { apply PWC_bound. }
    rewrite size_map -(size_map fst) -/(unzip1 (PWC_s f)) -/xfs.
    by rewrite -ltnS ltnNge -has_find Hhas. }
  apply PWC_step; split; [by apply next_ge_id|].
  rewrite /next -/xfs.
  set x0 := nth x xfs _.
  have Hx0 : has (eqr x0) xs.
  { apply /hasP.
    exists x0; [|by apply /eqrP].
    rewrite mem_merge mem_cat; apply /orP; left.
    by apply mem_nth; rewrite -has_find. }
  rewrite (_ : x0 = nth x xs (find (eqr x0) xs)); [|by apply /eqrP /nth_find].
  apply /RlebP /sorted_leq_nth.
  { exact: le_trans. }
  { move=> x'; apply /RleP /Rle_refl. }
  { by []. }
  { by rewrite /in_mem /=. }
  { by rewrite /in_mem /= -has_find. }
  apply sub_find => x' /eqrP <-.
  by apply nth_find. }
case_eq (has (Rleb x) xgs) => Hhas; last first.
{ apply Rle_antisym.
  { apply /PWC_non_decr /next_ge_id. }
  rewrite [in X in _ <= X]/PWC2f [in X in _ <= X]nth_default.
  { apply PWC_bound. }
  rewrite size_map -(size_map fst) -/(unzip1 (PWC_s g)) -/xgs.
  by rewrite -ltnS ltnNge -has_find Hhas. }
apply PWC_step; split; [by apply next_ge_id|].
rewrite /next -/xgs.
set x0 := nth x xgs _.
have Hx0 : has (eqr x0) xs.
{ apply /hasP.
  exists x0; [|by apply /eqrP].
  rewrite mem_merge mem_cat; apply /orP; right.
  by apply mem_nth; rewrite -has_find. }
rewrite (_ : x0 = nth x xs (find (eqr x0) xs)); [|by apply /eqrP /nth_find].
apply /RlebP /sorted_leq_nth.
{ exact: le_trans. }
{ move=> x'; apply /RleP /Rle_refl. }
{ by []. }
{ by rewrite /in_mem /=. }
{ by rewrite /in_mem /= -has_find. }
apply sub_find => x' /eqrP <-.
by apply nth_find.
Qed.

Lemma PWC_add_segment f g x x' :
  in_segment f x x' -> in_segment g x x' -> in_segment (PWC_add f g) x x'.
Proof.
rewrite /in_segment /PWC_add /= /unzip1 -map_comp map_id.
move=> /allP Hf /allP Hg.
apply /allP => t.
by rewrite mem_merge mem_cat => /orP [Ht|Ht]; [apply Hf|apply Hg].
Qed.

Lemma nuPWC (T : R+*) (C : R+) (J : R) :
  forall x x',
    { f : PWC |
      in_segment f x x'
      /\ forall t, x <= t <= x' -> PWC2f f t = C * IZR (Zceil ((t + J) / T))}.
Proof.
move=> x x'.
suff: { f : PWC |
    in_segment f x x'
    /\ all (Rleb (PWC2f f x)) (rcons (unzip2 (PWC_s f)) (PWC_c f))
    /\ forall t : R, x <= t <= x' -> PWC2f f t = C * IZR (Zceil ((t + J) / T)) }.
{ by move=> [f [H [_ H']]]; exists f. }
have PT := posR_pos T.
have nnT := nnR_pos T.
have PdT : 0 < / T; [by apply Rinv_0_lt_compat|].
have nndT : 0 <= / T; [lra|].
case (RleP x x')=> [|/Rnot_le_lt] Hxx'; last first.
{ exists (PWC_cst 0); split; [by []|split].
  { rewrite /= andbC /= /PWC2f /=; apply /RleP /Rle_refl. }
  move=> t [Hx Hx'].
  exfalso; apply (Rlt_irrefl x), (Rle_lt_trans _ x'); [|exact Hxx'].
  by move: Hx'; apply Rle_trans. }
set n := Z.to_nat (Zceil ((x' - x) / T)).
have Hn : x' - x <= INR n * T.
{ rewrite /n INR_IZR_INZ Z2Nat.id; last first.
  { apply le_IZR; refine (Rle_trans _ _ _ _ (Zceil_ub _)).
    by apply Rmult_le_pos; [lra|]. }
  apply (Rle_trans _ ((x' - x) / T * T)).
  { by rewrite Rmult_assoc Rinv_l ?Rmult_1_r; [right|lra]. }
  apply Rmult_le_compat_r; [apply nnR_pos|apply Zceil_ub]. }
move: n Hn => n.
elim: n x x' Hxx' => [|n IHn] x x' Hxx'.
{ rewrite Rmult_0_l => Hxx''.
  exists (PWC_cst (C * IZR (Zceil ((x + J) / T)))); split; [by []|split].
  { rewrite /= /PWC2f /= andbC /=; apply /RleP /Rle_refl. }
  by move=> t Ht; have -> : t = x; [lra|]. }
pose x'' := x + T.
pose x0 := IZR (Zceil ((x + J) / T)) * T - J.
pose y0 := C * IZR (Zceil ((x0 + J) / T)).
have Hx0 : IZR (Zceil ((x0 + J) / T)) = (x0 + J) / T.
{ rewrite /x0 (_ : (_ - J + J) / T = IZR (Zceil ((x + J) / T))); [|field; lra].
  by rewrite Zceil_IZR. }
have Hxx0 : x <= x0.
{ rewrite /x0; apply (Rplus_le_reg_r J); ring_simplify.
  apply (Rmult_le_reg_r (/ T)) => //.
  apply (Rle_trans _ _ _ (Zceil_ub _)); field_simplify; [by right|].
  apply Rgt_not_eq, posR_pos. }
have Hx0x'' : x0 < x''.
{ rewrite /x0 /x''.
  apply (Rplus_lt_reg_r J), (Rmult_lt_reg_r (/ T)) => //.
  rewrite (_ : (_ - J + J) * / T = IZR (Zceil ((x + J) / T))); [|field; lra].
  rewrite (_ : (x + T + J) * / T = (x + J) / T + 1); [|by field; lra].
  apply Zceil_lb. }
have Hlx0 : forall t, x <= t <= x0 -> y0 = C * IZR (Zceil ((t + J) / T)).
{ move=> t [Hxt Htx0].
  rewrite /y0 Hx0 /x0.
  rewrite (_ : (_ - J + J) / T = IZR (Zceil ((x + J) / T))); [|by field; lra].
  apply f_equal2 => //; apply f_equal, Zle_antisym.
  { apply Zceil_le, Rmult_le_compat_r => //.
    by apply Rplus_le_compat_r. }
  apply Zceil_glb.
  apply (Rmult_le_reg_r T _ _ PT).
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; [|lra].
  apply (Rplus_le_reg_r (- J)); ring_simplify; rewrite Rplus_comm.
  by rewrite -/(IZR (Zceil ((x + J) * / T)) * T - J) -/x0. }
case (RleP x' x0) => [|/Rnot_le_lt] Hx0x'.
{ move=> _.
  exists (PWC_cst y0); split; [by []|split].
  { rewrite /= /PWC2f /= andbC /=; apply /RleP /Rle_refl. }
  move=> t [Hxt Htx']; rewrite /PWC2f /=.
  apply (Hlx0 _ (conj Hxt (Rle_trans _ _ _ Htx' Hx0x'))). }
case (RleP x'' x') => [|/Rnot_le_lt] Hx''x' Hx'x; last first.
{ pose y0' := y0 + C.
  have Hy0y0' : y0 <= y0'; [by rewrite /y0'; move: (nnR_pos C); lra|].
  exists (PWC_singleton x0 Hy0y0'); split; [|split].
  { rewrite /in_segment /= andbC /=; apply /andP.
    by split; [apply /RleP|apply /RltP]. }
  { rewrite /=; apply /andP; split; [|apply /andP; split; [|by []]].
    { rewrite (PWC_singleton_before _ Hxx0); apply /RleP /Rle_refl. }
    by rewrite (PWC_singleton_before _ Hxx0); apply /RleP. }
  move=> t [Ht Ht'].
  case (Rle_or_lt t x0) => Htx0.
  { by rewrite (PWC_singleton_before _ Htx0); apply Hlx0. }
  rewrite (PWC_singleton_after _ Htx0) /y0' /y0.
  rewrite -{2}(Rmult_1_r C) -Rmult_plus_distr_l -plus_IZR.
  rewrite /x0 (_ : (_ - J + J) / T = IZR (Zceil ((x + J) / T))); [|by field; lra].
  rewrite Zceil_IZR.
  apply f_equal2 => //; apply f_equal; symmetry; apply Zceil_imp; split.
  { rewrite (_ : (_ - 1 = Zceil ((x + J) / T))%Z); [|by ring_simplify].
    apply (Rmult_lt_reg_r T _ _ PT).
    rewrite Rmult_assoc Rinv_l ?Rmult_1_r; [|by lra].
    apply (Rplus_lt_reg_r (- J)); ring_simplify; rewrite Rplus_comm.
    by rewrite Rplus_comm -/x0. }
  rewrite plus_IZR.
  apply (Rmult_le_reg_r T _ _ PT).
  rewrite Rmult_assoc Rinv_l ?Rmult_1_r; [|by lra].
  apply (Rle_trans _ (x'' + J)); [lra|].
  apply (Rplus_le_reg_r (- T)); rewrite /x''; ring_simplify; rewrite Rmult_comm.
  apply (Rmult_le_reg_r (/ T)); [by apply Rinv_0_lt_compat, posR_pos|].
  rewrite Rmult_assoc Rinv_r ?Rmult_1_r; [|by lra].
  apply Zceil_ub. }
have Hx'x'' : x' - x'' <= INR n * T.
{ rewrite /x''.
  apply (Rplus_le_reg_r T); ring_simplify.
  apply (Rle_trans _ _ _ Hx'x); rewrite S_INR; lra. }
move: (IHn _ _ Hx''x' Hx'x'') => {IHn} [[s c Hs1 Hs2] /= [Hx0s [Hfx'' Hf]]].
pose s' := (x0, y0) :: s.
have Hy0fx'' : y0 <= PWC2f (Build_PWC Hs1 Hs2) x''.
{ rewrite /y0 (Hf x'' (conj (Rle_refl _) Hx''x')).
  apply Rmult_le_compat_l; [by apply nnR_pos|].
  apply IZR_le, Zceil_le, Rmult_le_compat_r => //; lra. }
have Hs'1 : sorted Rleb (unzip1 s').
{ rewrite /= path_min_sorted //.
  move: Hx0s; apply sub_all => y /andP [] /RleP Hx''y _; apply /RleP; lra. }
have Hs'2 : sorted Rleb (rcons (unzip2 s') c).
{ rewrite /= path_min_sorted //.
  move: Hfx''; apply sub_all => y /RleP Hfx''; apply /RleP.
  by move: Hfx''; apply Rle_trans. }
exists (Build_PWC Hs'1 Hs'2); split; [|split].
{ rewrite /in_segment /=; apply /andP; split.
  { by apply /andP; split; [apply /RleP|apply /RltP]. }
  move: Hx0s; rewrite /in_segment; apply sub_all => y /andP [] /RleP Hx''y /RltP Hyx'.
  apply /andP; split; [apply /RleP|apply /RltP]; lra. }
{ rewrite /= /PWC2f /=; move: (Hxx0) => /RlebP -> /=; apply /andP; split.
  { apply /RleP /Rle_refl. }
  move: Hfx''; apply sub_all => y /RleP Hfx''y; apply /RleP.
  by move: Hfx''y; apply Rle_trans. }
move=> t [Hxt Htx'].
case (Rle_or_lt t x'') => Htx''; last first.
{ rewrite -(Hf t (conj (Rlt_le _ _ Htx'') Htx')).
  move: (Rle_lt_trans _ _ _ (Rlt_le _ _ Hx0x'') Htx'').
  rewrite /PWC2f /=.
  by move=> /Rlt_not_le /RlebP /negbTE ->. }
case (Rle_or_lt t x0) => Htx0.
{ by rewrite /PWC2f /=; move: (Htx0) => /RlebP -> /=; apply Hlx0. }
rewrite /PWC2f /=; move: (Htx0) => /Rlt_not_le /RlebP /negbTE -> /=.
rewrite -(@eq_in_find _ (Rleb x'')); last first.
{ move=> y Hy.
  move: Hx0s => /allP H; move: (H y Hy) => {H} /andP [Hx''y _].
  rewrite Hx''y; symmetry; apply /RleP; move: Hx''y => /RleP; lra. }
rewrite -/(PWC2f (Build_PWC Hs1 Hs2) x'').
rewrite (Hf x'' (conj (Rle_refl _) Hx''x')).
apply f_equal2 => //; apply f_equal.
symmetry; apply Zceil_imp; split; last first.
{ apply (Rle_trans _ _ _ (Zceil_ub _)), IZR_le, Zceil_le.
  by apply Rmult_le_compat_r => //; apply Rplus_le_compat_r. }
rewrite minus_IZR; apply (Rplus_lt_reg_r 1); ring_simplify.
apply (Rle_lt_trans _ (IZR (Zceil ((x0 + T + J) / T)))).
{ apply IZR_le, Zceil_le, Rmult_le_compat_r => //; rewrite /x''; lra. }
rewrite /x0.
rewrite (_ : (_ + J) / T = IZR (Zceil ((x + J) / T)) + 1); [|by field; lra].
rewrite -plus_IZR Zceil_IZR plus_IZR; apply Rplus_lt_compat_r.
apply (Rle_lt_trans _ (IZR (Zceil ((x0 + J) / T)))).
{ apply IZR_le, Zceil_le, Rmult_le_compat_r => //; lra. }
rewrite Hx0; apply Rmult_lt_compat_r; lra.
Qed.
