(** Definition of non negative and positive real numbers.

   The basic idea is to define them as a real number x plus a proof of x >= 0.
   Unfortunately such proofs are not unique, meaning we could have multiple
   representations for a same value. To avoid this, the proof is encoded
   using a boolean predicate (to keep it short, all proofs then reduce to
   true = true) using the eqtype mechanism from Ssreflect. *)

Require Import mathcomp.ssreflect.ssreflect.
Require Import mathcomp.ssreflect.ssrbool.
Require Import mathcomp.ssreflect.eqtype.
Require Import Reals.
Require Import mathcomp.analysis.Rstruct.
Require Import Psatz.

Local Open Scope R_scope.

(** * Non negative reals *)

(** nnR is the subset of non negative values of R. *)
Record nnR := { nnR_val :> R; nnR_prop : Rleb 0 nnR_val }.

(** It then inherits the eqType structure of R. *)
Canonical nnR_subType := [subType for nnR_val].
Definition nnR_eqMixin := Eval hnf in [eqMixin of nnR by <:].
Canonical nnR_eqType := Eval hnf in EqType nnR nnR_eqMixin.

Lemma nnR_pos (n : nnR) : 0 <= n. Proof. apply /RleP /valP. Qed.

Definition mk_nnR (x : R) (Px : 0 <= x) :=
  Build_nnR x (introTF (c:=true) (RleP 0 x) Px).

(** Notation *)

Notation "R+" := (nnR).

(** * Positive reals *)

(** posR is the subset of non zero values of nnR *)

Record posR := { posR_val :> nnR; posR_prop : ~~ eqr posR_val 0 }.

(** It then inherits the eqType structure of nnR. *)
Canonical posR_subType := [subType for posR_val].
Definition posR_eqMixin := Eval hnf in [eqMixin of posR by <:].
Canonical posR_eqType := Eval hnf in EqType posR posR_eqMixin.

Lemma posR_pos (n : posR) : 0 < n.
Proof.
by move: (nnR_pos n) => [//|]; case n => {}n /eqP H /= H'; exfalso; apply H.
Qed.

Definition mk_posR (x : R) (Px : 0 < x) :=
  Build_posR
    (mk_nnR x (Rlt_le 0 x Px))
    (introNTF (c:=true) eqP (Rgt_not_eq x 0 Px)).

(** Notation *)

Notation "R+*" := (posR).

(** Convenient tactics *)

Ltac lra' :=
  match goal with
  | |- 0 <= / _ => by apply Rlt_le, Rinv_0_lt_compat, posR_pos
  | |- 0 < / _ => by apply Rinv_0_lt_compat, posR_pos
  | |- _ <> 0 => by apply Rgt_not_eq, posR_pos
  | |- 0 <= _ => by apply nnR_pos
  | |- 0 < _ => by apply posR_pos
  | _ => lra
  end.

Ltac field' := field; (try split); lra'.
